/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-
 *
 * tracker-tag-cloud-model.c: Model for the TrackerTagCloud widget
 *
 * Copyright (C) 2007 Marko Anastasov <marko@marko.anastasov.name>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <math.h>
#include "tracker-tag-cloud-model.h"

typedef struct _CloudStats CloudStats;
typedef struct _TagData TagData;

enum {
	CHANGED,
	LAST_SIGNAL
};

struct _CloudStats
{
	int font_range;

	int min_tag_count;
	int max_tag_count;

	double min_log;
	double max_log;
	double log_range;
};

struct _TagData
{
    int count;
    int font_size;
};

struct _TrackerTagCloudModelPrivate
{
	CloudStats *stats;

	GHashTable *tags;

	gint min_font_size;
	gint max_font_size;
};

GType tracker_tag_cloud_model_get_type ();
static void tracker_tag_cloud_model_dispose (GObject *object);
static void tracker_tag_cloud_model_finalize (GObject *object);
static void tracker_tag_cloud_model_install_signals (GObjectClass *gobject_class);
static void cloud_stats_init (CloudStats *stats);

static guint tag_cloud_model_signals[LAST_SIGNAL] = { 0 };

static GObjectClass *parent_class = NULL;

/*** GObject stuff ***/

static void
tracker_tag_cloud_model_class_init (TrackerTagCloudModelClass *klass)
{
	GObjectClass *gobject_class = (GObjectClass *) klass;

	parent_class = g_type_class_peek_parent (klass);

	gobject_class->dispose  = tracker_tag_cloud_model_dispose;
	gobject_class->finalize = tracker_tag_cloud_model_finalize;

	tracker_tag_cloud_model_install_signals (gobject_class);

	klass->changed = NULL;
}

static void
tracker_tag_cloud_model_install_signals (GObjectClass *gobject_class)
{
	tag_cloud_model_signals[CHANGED] =
		g_signal_new ("changed",
			      G_TYPE_FROM_CLASS (gobject_class),
			      G_SIGNAL_RUN_FIRST,
			      G_STRUCT_OFFSET (TrackerTagCloudModelClass,
					       changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
}

static void
tracker_tag_cloud_model_init (TrackerTagCloudModel *model,
			      TrackerTagCloudModelClass *class)
{
	g_return_if_fail (TRACKER_IS_TAG_CLOUD_MODEL (model));

	/* allocate private data */
	model->priv = g_new0 (TrackerTagCloudModelPrivate, 1);

	model->priv->stats = g_new0 (CloudStats, 1);
	cloud_stats_init (model->priv->stats);

	model->priv->tags = g_hash_table_new_full (g_str_hash, g_str_equal,
						   g_free, g_free);
}

static void
tracker_tag_cloud_model_dispose (GObject *object)
{
	/* TrackerTagCloudModel *model = (TrackerTagCloudModel *) object; */

	G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
tracker_tag_cloud_model_finalize (GObject *object)
{
	TrackerTagCloudModel *model = (TrackerTagCloudModel *) object;

	g_free (model->priv->stats);
	model->priv->stats = NULL;

	g_hash_table_destroy (model->priv->tags);
	model->priv->tags = NULL;

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

GType
tracker_tag_cloud_model_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static const GTypeInfo info = {
			sizeof (TrackerTagCloudModelClass),
			(GBaseInitFunc) NULL,       /* base_init */
			(GBaseFinalizeFunc) NULL,   /* base_finalize */
			(GClassInitFunc) tracker_tag_cloud_model_class_init,
			NULL,   /* class_finalize */
			NULL,   /* class_data */
			sizeof (TrackerTagCloudModel),
			0,      /* n_preallocs */
			(GInstanceInitFunc) tracker_tag_cloud_model_init
		};

		type = g_type_register_static (G_TYPE_OBJECT,
					       "TrackerTagCloudModel",
					       &info, 0);
	}

	return type;
}

/*** Implementation ***/

static void
cloud_stats_init (CloudStats *stats)
{
	if (! stats) return;

	stats->font_range = 0;
	stats->min_tag_count = 0;
	stats->max_tag_count = 0;
	stats->min_log = 0;
	stats->max_log = 0;
	stats->log_range = 1;
}

/* The following two functions have been copied from tracker-indexer.c */

static gint
prepend_key_pointer (gpointer key, gpointer value, gpointer data)
{
  	GSList **plist = data;
  	*plist = g_slist_prepend (*plist, key);
  	return 1;
}


static GSList * 
g_hash_table_key_slist (GHashTable *table)
{
  	GSList *rv = NULL;
  	g_hash_table_foreach (table, (GHFunc) prepend_key_pointer, &rv);
  	return rv;
}

typedef struct
{
	int min;
	int max;
} TagCountRecord;

static void
find_min_max_tag_count (const gchar *key,
			const TagData *tag_data,
			TagCountRecord *count_record)
{
	if (tag_data->count < count_record->min) {
		count_record->min = tag_data->count;
		return;
	}

	if (tag_data->count > count_record->max)
		count_record->max = tag_data->count;
}

static void
calculate_stats (TrackerTagCloudModel *model)
{
	CloudStats *stats;
	GSList *keys = NULL;
	TagCountRecord *count_record = NULL;
	gchar *first_key;
	TagData *first_tag_data_item;

	stats = model->priv->stats;

	stats->font_range =
		model->priv->max_font_size - model->priv->min_font_size;

	if (g_hash_table_size (model->priv->tags) == 0) return;

	/* Find the min and max tag count values */
	keys = g_hash_table_key_slist (model->priv->tags);
	count_record = g_new0 (TagCountRecord, 1);

	first_key = (gchar *) g_slist_nth_data (keys, 0);

	first_tag_data_item = (TagData *)
		g_hash_table_lookup (model->priv->tags, first_key);
	count_record->min = first_tag_data_item->count;
	count_record->max = first_tag_data_item->count;

	g_hash_table_foreach (model->priv->tags,
			      (GHFunc) find_min_max_tag_count,
			      count_record);

	/* Update stats */
	stats->min_tag_count = count_record->min;
	stats->max_tag_count = count_record->max;

	stats->min_log = log(count_record->min);
	stats->max_log = log(count_record->max);

	if (stats->max_log == stats->min_log)
		stats->log_range = 1;
	else
		stats->log_range = stats->max_log - stats->min_log;

	g_free (count_record);
}

static void
calculate_font_size (const gchar *key,
		     TagData *tag_data,
		     TrackerTagCloudModel *model)
{
	CloudStats *stats;
	gdouble count_ratio;

	stats = model->priv->stats;

	count_ratio =
		(log(tag_data->count) - stats->min_log) / stats->log_range;

	tag_data->font_size =
		model->priv->min_font_size +
		(gint) (stats->font_range * count_ratio);
}

static void
update_font_sizes (TrackerTagCloudModel *model)
{
	g_hash_table_foreach (model->priv->tags,
			      (GHFunc) calculate_font_size,
			      model);
}

static void
update_data_and_notify (TrackerTagCloudModel *model)
{
	calculate_stats (model);
	update_font_sizes (model);

	g_signal_emit (model, tag_cloud_model_signals [CHANGED], 0);
}

TrackerTagCloudModel *
tracker_tag_cloud_model_new (gint min_font_size, gint max_font_size)
{
	TrackerTagCloudModel *model;

	model = g_object_new (TYPE_TRACKER_TAG_CLOUD_MODEL, NULL);

	model->priv->min_font_size = min_font_size;
	model->priv->max_font_size = max_font_size;
  
	return model;
}

void
tracker_tag_cloud_model_add_tag (TrackerTagCloudModel* model, const gchar* tag)
{
	TagData* tag_data = NULL;
	gchar *tag_str;

	g_return_if_fail (TRACKER_IS_TAG_CLOUD_MODEL (model));

	if ((tag_data = g_hash_table_lookup (model->priv->tags, tag)) == NULL) {
		tag_data = g_new0 (TagData, 1);
		tag_data->count = 1;
		tag_str = g_strdup (tag);
		g_hash_table_insert (model->priv->tags, tag_str, tag_data);
	} else {
		tag_data->count++;
	}

	update_data_and_notify (model);
}

static void
load_font_size_hash (const gchar *key,
		     const TagData *tag_data,
		     GHashTable *hash)
{
	gchar *key_copy = NULL;
	int *count = NULL;

	key_copy = g_strdup (key);

	count = g_new0 (int, 1);
	*count = tag_data->font_size;

	g_hash_table_insert (hash, key_copy, count);
}

GHashTable *
tracker_tag_cloud_model_get_font_sizes (TrackerTagCloudModel* model)
{
	GHashTable *hash = NULL;

	g_return_val_if_fail (TRACKER_IS_TAG_CLOUD_MODEL (model), NULL);

	hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

	g_hash_table_foreach (model->priv->tags,
			      (GHFunc) load_font_size_hash,
			      hash);

	return hash;
}

gint
tracker_tag_cloud_model_get_min_font_size (TrackerTagCloudModel* model)
{
	g_return_val_if_fail (TRACKER_IS_TAG_CLOUD_MODEL (model), 0);

	return model->priv->min_font_size;
}

void
tracker_tag_cloud_model_set_min_font_size (TrackerTagCloudModel* model,
					   gint value)
{
	g_return_if_fail (TRACKER_IS_TAG_CLOUD_MODEL (model));
	g_return_if_fail (value > 0);

	model->priv->min_font_size = value;

	update_data_and_notify (model);
}

gint tracker_tag_cloud_model_get_max_font_size (TrackerTagCloudModel* model)
{
	g_return_val_if_fail (TRACKER_IS_TAG_CLOUD_MODEL (model), 0);

	return model->priv->max_font_size;
}

void
tracker_tag_cloud_model_set_max_font_size (TrackerTagCloudModel* model,
					   gint value)
{
	g_return_if_fail (TRACKER_IS_TAG_CLOUD_MODEL (model));
	g_return_if_fail (value > 0);

	model->priv->max_font_size = value;

	update_data_and_notify (model);
}
