/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-
 *
 * tracker-tag-cloud.c: Tracker tag cloud widget
 *
 * Copyright (C) 2007 Marko Anastasov <marko@marko.anastasov.name>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <goocanvas.h>
#include "tracker-tag-cloud.h"

#define TRACKER_TAG_CLOUD_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), TRACKER_TYPE_TAG_CLOUD, TrackerTagCloudPrivate))


#define TAG_SPACING 8
#define TAG_BORDER_MIN 5
#define TAG_BORDER_MAX (TAG_BORDER_MIN * 4)

enum FILL_COLOUR
{
	FILL_COLOUR_BLUE       = 0x0000ffff,
	FILL_COLOUR_LIGHT_BLUE = 0x6060ffff,
	FILL_COLOUR_RED        = 0xff0000ff
};

/* We want to give the canvas some space smaller then what the scrolled
 * window has been allocated, so that the scroll bars do not show up by default.
 */
#define CANVAS_OFFSET 30

typedef struct _TagItem TagItem;
typedef struct _ButtonPressCbData ButtonPressCbData;

enum
{
	PROP_0,
	PROP_RANDOMIZE_BORDERS,
	PROP_UNDERLINE_TAGS,
	PROP_LIGHTEN_RARE_TAGS
};

enum
{
	CLICKED,
	LAST_SIGNAL
};

struct _TagItem
{
	GooCanvasItem *text_item;
	gchar *tag;
	guint fill_colour;
};

struct _TrackerTagCloudPrivate
{
	GSList *tag_items;
	GooCanvasItem *root_item;

	TrackerTagCloudModel *model;

	GRand *grand;
	gboolean randomize_borders;

	gboolean underline_tags;
	gboolean lighten_rare_tags;

	int alloc_width;
	int alloc_height;

	GSList *button_press_data;
};

static void tracker_tag_cloud_finalize (GObject *obj);
static void tracker_tag_cloud_init (TrackerTagCloud *cloud);
static void tracker_tag_cloud_install_signals (GObjectClass *gobject_class);
static void tracker_tag_cloud_install_properties (GObjectClass *gobject_class);
static void tracker_tag_cloud_get_property (GObject *object,
					    guint prop_id,
					    GValue *value,
					    GParamSpec *pspec);
static void tracker_tag_cloud_set_property (GObject *object,
					    guint prop_id,
					    const GValue *value,
					    GParamSpec *pspec);
static void tracker_tag_cloud_size_allocate (GtkWidget *widget,
					     GtkAllocation *allocation);
static void free_tag_items_slist (GSList **list);
static void free_button_press_cb_data (GSList **list);
static void draw_tags (TrackerTagCloud *cloud);

static GtkScrolledWindowClass *parent_class = NULL;
static guint tag_cloud_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(TrackerTagCloud, tracker_tag_cloud, GTK_TYPE_SCROLLED_WINDOW);

static void
tracker_tag_cloud_class_init (TrackerTagCloudClass *klass)
{
	GObjectClass   *gobject_class = G_OBJECT_CLASS (klass);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	gobject_class->finalize = tracker_tag_cloud_finalize;

	widget_class->size_allocate = tracker_tag_cloud_size_allocate;

	gobject_class->set_property = tracker_tag_cloud_set_property;
	gobject_class->get_property = tracker_tag_cloud_get_property;

	g_type_class_add_private (klass, sizeof (TrackerTagCloudPrivate));

	tracker_tag_cloud_install_signals (gobject_class);
	tracker_tag_cloud_install_properties (gobject_class);
}

static void
tracker_tag_cloud_install_signals (GObjectClass *gobject_class)
{
	tag_cloud_signals[CLICKED] =
		g_signal_new ("clicked",
			      G_TYPE_FROM_CLASS (gobject_class),
			      G_SIGNAL_RUN_FIRST,
			      G_STRUCT_OFFSET (TrackerTagCloudClass,
					       clicked),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__POINTER,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_POINTER);
}

static void
tracker_tag_cloud_install_properties (GObjectClass *gobject_class)
{
	g_object_class_install_property (gobject_class,
					 PROP_RANDOMIZE_BORDERS,
					 g_param_spec_boolean (
						 "randomize-borders",
						 "Randomize borders",
						 "Whether the borders of tag lines are of random length",
						 TRUE,
						 G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class,
					 PROP_UNDERLINE_TAGS,
					 g_param_spec_boolean (
						 "underline-tags",
						 "Underline tags",
						 "Whether to display the tags underlined",
						 FALSE,
						 G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class,
					 PROP_LIGHTEN_RARE_TAGS,
					 g_param_spec_boolean (
						 "lighten-rare-tags",
						 "Lighten rare tags",
						 "Whether to draw the tags of half the largest font size or less in a lighter shade",
						 TRUE,
						 G_PARAM_READWRITE));
}

static void
tracker_tag_cloud_finalize (GObject *obj)
{
	TrackerTagCloud *cloud = (TrackerTagCloud *) obj;
	TrackerTagCloudPrivate *priv = TRACKER_TAG_CLOUD_GET_PRIVATE (cloud);

	free_tag_items_slist (&(priv->tag_items));
	g_rand_free (priv->grand);
	free_button_press_cb_data (&(priv->button_press_data));

	if (G_OBJECT_CLASS(parent_class)->finalize != NULL)
		G_OBJECT_CLASS(parent_class)->finalize(obj);
}

static void
tracker_tag_cloud_init (TrackerTagCloud *cloud)
{
	TrackerTagCloudPrivate *priv = TRACKER_TAG_CLOUD_GET_PRIVATE (cloud);

	cloud->canvas = NULL;

	priv->tag_items = NULL;
	priv->root_item = NULL;

	priv->grand = g_rand_new ();

	priv->randomize_borders = TRUE;
	priv->underline_tags = FALSE;
	priv->lighten_rare_tags = TRUE;

	priv->alloc_width = 0;
	priv->alloc_height = 0;

	priv->button_press_data = NULL;

	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (cloud),
					     GTK_SHADOW_IN);

	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (cloud),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
}

static void
tracker_tag_cloud_get_property (GObject *object,
				guint prop_id,
				GValue *value,
				GParamSpec *pspec)
{
	TrackerTagCloud *cloud;
	TrackerTagCloudPrivate *priv;
  
	cloud = TRACKER_TAG_CLOUD (object);
	priv = TRACKER_TAG_CLOUD_GET_PRIVATE (cloud);

	switch (prop_id) {
	case PROP_RANDOMIZE_BORDERS:
		g_value_set_boolean (value, priv->randomize_borders);
		break;
	case PROP_UNDERLINE_TAGS:
		g_value_set_boolean (value, priv->underline_tags);
		break;
	case PROP_LIGHTEN_RARE_TAGS:
		g_value_set_boolean (value, priv->lighten_rare_tags);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
tracker_tag_cloud_set_property (GObject *object,
				guint prop_id,
				const GValue *value,
				GParamSpec *pspec)
{
	TrackerTagCloud *cloud;
	TrackerTagCloudPrivate *priv;
  
	cloud = TRACKER_TAG_CLOUD (object);
	priv = TRACKER_TAG_CLOUD_GET_PRIVATE (cloud);

	switch (prop_id) {
	case PROP_RANDOMIZE_BORDERS:
		priv->randomize_borders = g_value_get_boolean (value);
		break;
	case PROP_UNDERLINE_TAGS:
		priv->underline_tags = g_value_get_boolean (value);
		break;
	case PROP_LIGHTEN_RARE_TAGS:
		priv->lighten_rare_tags = g_value_get_boolean (value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

GtkWidget *
tracker_tag_cloud_new ()
{
	GtkWidget *widget;

	widget = g_object_new (TRACKER_TYPE_TAG_CLOUD, NULL);

	return widget;
}

GtkWidget *
tracker_tag_cloud_new_with_model (TrackerTagCloudModel *model)
{
	GtkWidget *widget;

	widget = g_object_new (TRACKER_TYPE_TAG_CLOUD, NULL);
	tracker_tag_cloud_set_model (TRACKER_TAG_CLOUD (widget), model);

	return widget;
}

static void
tracker_tag_cloud_size_allocate (GtkWidget *widget, GtkAllocation *alloc)
{
	TrackerTagCloud *cloud = (TrackerTagCloud *) widget;
	TrackerTagCloudPrivate *priv = TRACKER_TAG_CLOUD_GET_PRIVATE (cloud);
	gboolean redraw;

	if (! cloud->canvas) {
		cloud->canvas = goo_canvas_new();

		gtk_widget_show (GTK_WIDGET (cloud->canvas));
		gtk_container_add (GTK_CONTAINER (cloud), cloud->canvas);

		priv->root_item =
			goo_canvas_get_root_item (GOO_CANVAS (cloud->canvas));
	}

	/* We need to redraw (and re-order) the tags if the widget
	 * has been resized. 
	 */

	redraw = ((priv->alloc_width != alloc->width) ||
		  (priv->alloc_height != alloc->height));

	if (! redraw) {
		GTK_WIDGET_CLASS(parent_class)->size_allocate (widget, alloc);
		return;
	}

	priv->alloc_width = alloc->width;
	priv->alloc_height = alloc->height;

	draw_tags (cloud);

	GTK_WIDGET_CLASS(parent_class)->size_allocate (widget, alloc);
}

static void
model_changed_cb (TrackerTagCloudModel *model, TrackerTagCloud *cloud)
{
	draw_tags (cloud);
}

void
tracker_tag_cloud_set_model (TrackerTagCloud *cloud,
			     TrackerTagCloudModel *model)
{
	g_return_if_fail (TRACKER_IS_TAG_CLOUD (cloud));
	g_return_if_fail (TRACKER_IS_TAG_CLOUD_MODEL (model));

	TrackerTagCloudPrivate *priv = TRACKER_TAG_CLOUD_GET_PRIVATE (cloud);

	priv->model = model;

	if (cloud->canvas) draw_tags (cloud);

	g_signal_connect (model,
			  "changed",
			  (GCallback) model_changed_cb,
			  cloud);
}

/* Struct carrying information for the tag line drawing */
typedef struct
{
	TrackerTagCloud *cloud;

	gdouble canvas_left;
	gdouble canvas_top;
	gdouble canvas_right;
	gdouble canvas_bottom;
	gint    visible_width;
	gdouble line_height;

	gint    bold_colour_limit;
	gint    current_line;
	gint    tags_in_current_line;
	gdouble text_item_x;
	gdouble text_item_y;
	gint    last_border;
} TagLineData;

static TagLineData *
tag_line_data_new (TrackerTagCloud *cloud)
{
	TagLineData *data;

	data = g_new0 (TagLineData, 1);

	data->cloud = cloud;
	data->bold_colour_limit = 0;
	data->current_line = 0;
	data->tags_in_current_line = 0;
	data->text_item_x = TAG_BORDER_MIN;
	data->text_item_y = 0;
	data->last_border = TAG_BORDER_MIN;

	return data;
}

struct _ButtonPressCbData
{
	TrackerTagCloud *cloud;
	TagItem *tag_item;
};

static ButtonPressCbData *
button_press_cb_data_new (TrackerTagCloud *cloud, TagItem *tag_item)
{
	ButtonPressCbData *data;

	data = g_new0 (ButtonPressCbData, 1);
	data->cloud = cloud;
	data->tag_item = tag_item;

	return data;
}

static void
free_button_press_cb_data (GSList **list)
{
	if (list == NULL) return;

	GSList *it;

	for (it = *list; it; it = it->next) {
		ButtonPressCbData *bp_data = (ButtonPressCbData *) it->data;
		g_free (bp_data);
	}

	g_slist_free (*list);
}

static gboolean
button_press_cb (GooCanvasItem* item,
		 GooCanvasItem* target,
		 GdkEventButton* event,
		 ButtonPressCbData* data)
{
	g_signal_emit (data->cloud,
		       tag_cloud_signals [CLICKED],
		       0,
		       data->tag_item->tag);

	return TRUE;
}

static gboolean
enter_notify_cb (GooCanvasItem* item,
		 GooCanvasItem* target,
		 GdkEventMotion* event,
		 TagItem* tag_item)
{
	g_object_set (item, "fill_color_rgba", FILL_COLOUR_RED, NULL);

	return FALSE;
}

static gboolean
leave_notify_cb (GooCanvasItem* item,
		 GooCanvasItem* target,
		 GdkEventMotion* event,
		 TagItem* tag_item)
{
	g_object_set (item, "fill_color_rgba", tag_item->fill_colour, NULL);

	return FALSE;
}

static void
create_lines (const gchar *tag, int *font_size, TagLineData *data)
{
	gchar *tag_font_desc;
	gchar *tag_text;
	guint fill_colour = FILL_COLOUR_BLUE;

	GooCanvasItem *text_item;

	GooCanvasBounds bounds;
	gint text_item_width;

	TrackerTagCloud *cloud = data->cloud;
	TrackerTagCloudPrivate *priv = TRACKER_TAG_CLOUD_GET_PRIVATE (cloud);

	TagItem *new_tag_item;
	gboolean start_new_line;

	tag_font_desc = g_strdup_printf ("Bitstream vera sans %d", *font_size);

	if (priv->underline_tags)
		tag_text = g_strdup_printf ("<u> %s </u>", tag);
	else
		tag_text = g_strdup (tag);

	data->text_item_y = (data->current_line + 1) * data->line_height;

	if (priv->lighten_rare_tags &&
	    (*font_size <= data->bold_colour_limit))
		fill_colour = FILL_COLOUR_LIGHT_BLUE;

	text_item = goo_canvas_text_new (priv->root_item,
					 tag_text,
					 data->text_item_x,
					 data->text_item_y,
					 -1,
					 GTK_ANCHOR_WEST,
					 "fill-color-rgba", fill_colour,
					 "font", tag_font_desc,
					 "use-markup", TRUE,
					 NULL);

	g_free (tag_text);
	g_free (tag_font_desc);

        /* Determine the width of the created text item, and will it fit in the
	 * current line. We start a new line when we have at least one tag
	 * in the current, and the new one is too large for the available
	 * visible space in the scrolled window.
	 */
	goo_canvas_item_get_bounds (text_item, &bounds);
	text_item_width = (int) (bounds.x2 - bounds.x1);
	data->text_item_x += text_item_width;

	start_new_line = (((data->text_item_x + data->last_border) >
			   data->visible_width) &&
			  (data->tags_in_current_line >= 1));

	if (! start_new_line)
		data->text_item_x += TAG_SPACING;
	else {
		/* Set the coordinates to the beginning of a new line. */
		++data->current_line;

		if (priv->randomize_borders) {
			data->last_border = g_rand_int_range (priv->grand,
							      TAG_BORDER_MIN,
							      TAG_BORDER_MAX);
		}

		data->text_item_x = data->last_border;
		data->text_item_y =
			(data->current_line + 1) * data->line_height;

		data->tags_in_current_line = 0;

		/* If the canvas is too small vertically or horizontally
		 * to accept a new text item, expand it accordingly.
		 */
		if (data->text_item_y >
		    (data->canvas_bottom - data->canvas_top)) {
			/* Expand on the bottom */
			data->canvas_bottom += data->line_height;

			goo_canvas_set_bounds (GOO_CANVAS (cloud->canvas),
					       data->canvas_left,
					       data->canvas_top,
					       data->canvas_right,
					       data->canvas_bottom);
		}

		if (text_item_width >
		    (data->canvas_right - data->canvas_left)) {
			/* Expand on the right */
			data->canvas_right = text_item_width +
				(TAG_SPACING + data->last_border);

			goo_canvas_set_bounds (GOO_CANVAS (cloud->canvas),
					       data->canvas_left,
					       data->canvas_top,
					       data->canvas_right,
					       data->canvas_bottom);
		}

		/* Move the item to the beginning of the new line. */
		g_object_set (GOO_CANVAS_ITEM (text_item),
			      "x", data->text_item_x,
			      "y", data->text_item_y,
			      NULL);

		data->text_item_x += text_item_width + TAG_SPACING;
	}

	++(data->tags_in_current_line);

	/* Add the new goo text item to cloud's inventory. */
	new_tag_item = g_new0 (TagItem, 1);

	new_tag_item->text_item = text_item;
	new_tag_item->tag = g_strdup (tag);
	new_tag_item->fill_colour = fill_colour;

	priv->tag_items = g_slist_append (priv->tag_items, new_tag_item);

	/* Connect to text item's signals. */
	ButtonPressCbData *bp_data =
		button_press_cb_data_new (cloud, new_tag_item);

	priv->button_press_data =
		g_slist_append (priv->button_press_data, bp_data);

	g_signal_connect (text_item, "button_press_event",
			  (GtkSignalFunc) button_press_cb, bp_data);
	
	g_signal_connect (text_item, "enter_notify_event",
			  (GtkSignalFunc) enter_notify_cb, new_tag_item);

	g_signal_connect (text_item, "leave_notify_event",
			  (GtkSignalFunc) leave_notify_cb, new_tag_item);
}

static void
set_tags (TrackerTagCloud *cloud, GHashTable *tags)
{
	TrackerTagCloudPrivate *priv = TRACKER_TAG_CLOUD_GET_PRIVATE (cloud);
	TrackerTagCloudModel *model = priv->model;

	TagLineData *tag_line_data;

	gint min_font_size;
	gint max_font_size;

	if (g_hash_table_size (tags) == 0) return;

	tag_line_data = tag_line_data_new (cloud);

	goo_canvas_get_bounds (GOO_CANVAS (cloud->canvas),
			       &(tag_line_data->canvas_left),
			       &(tag_line_data->canvas_top),
			       &(tag_line_data->canvas_right),
			       &(tag_line_data->canvas_bottom));

	tag_line_data->visible_width = (int)
		(tag_line_data->canvas_right - tag_line_data->canvas_left);

	min_font_size = tracker_tag_cloud_model_get_min_font_size (model);
	max_font_size = tracker_tag_cloud_model_get_max_font_size (model);

	tag_line_data->line_height = max_font_size * 1.5;

	if (priv->lighten_rare_tags) {
		tag_line_data->bold_colour_limit =
			(min_font_size + (max_font_size - min_font_size) / 2);
	}

	g_hash_table_foreach (tags, (GHFunc) create_lines, tag_line_data);

	g_free (tag_line_data);
}

static void
free_tag_items_slist (GSList **tag_items)
{
	if (tag_items == NULL) return;

	GSList *it;

	for (it = *tag_items; it; it = it->next) {
		TagItem *olditem = (TagItem *) it->data;

		if (G_IS_OBJECT (olditem->text_item)) {
			goo_canvas_item_remove (
				GOO_CANVAS_ITEM (olditem->text_item));
		}

		g_free (olditem->tag);
		g_free (olditem);
	}

	g_slist_free (*tag_items);
	*tag_items = NULL;
}

static void
draw_tags (TrackerTagCloud *cloud)
{
	TrackerTagCloudPrivate *priv = TRACKER_TAG_CLOUD_GET_PRIVATE (cloud);
	GHashTable *tags;

	/* Remove the text items from canvas, create a new list */
	free_tag_items_slist (&(priv->tag_items));

	/* Set canvas bounds based on scrolled window's allocation */
	gtk_widget_set_size_request (cloud->canvas,
				     priv->alloc_width - CANVAS_OFFSET,
				     priv->alloc_height - CANVAS_OFFSET);

	goo_canvas_set_bounds (GOO_CANVAS (cloud->canvas),
			       0, 0,
			       priv->alloc_width - CANVAS_OFFSET,
			       priv->alloc_height - CANVAS_OFFSET);

	/* Redraw tag items on the canvas */
	tags = tracker_tag_cloud_model_get_font_sizes (priv->model);
	set_tags (cloud, tags);
	g_hash_table_destroy (tags);
}
